@@ -1,6 +1,6 @@
-/* $Id: DBIXS.h,v 11.3 2002/02/05 02:12:25 timbo Exp $
+/* $Id: DBIXS.h,v 11.11 2002/12/01 22:34:29 timbo Exp $
  *
- * Copyright (c) 1994, 1995, 1996, 1997, 1998, 1999  Tim Bunce  England
+ * Copyright (c) 1994-2002  Tim Bunce  Ireland
  *
  * See COPYRIGHT section in DBI.pm for usage and distribution rights.
  */
@@ -49,22 +49,19 @@ error You_need_to_upgrade_your_DBI_module_before_building_this_driver
 #endif
 
 
-#if defined(USE_THREADS) && !defined(DBI_NO_THREADS)
-#define DBI_USE_THREADS
-#define DBI_LOCK	MUTEX_LOCK(DBIS->mutex)
-#define DBI_UNLOCK	MUTEX_UNLOCK(DBIS->mutex)
-#define dbi_mutex perl_mutex
-#define dbi_cond  perl_cond
-#else
 #define DBI_LOCK
 #define DBI_UNLOCK
-#define dbi_mutex void
-#define dbi_cond  void
-#endif
+
+#ifndef DBI_NO_THREADS
+#ifdef USE_ITHREADS
+#define DBI_USE_THREADS
+#endif /* USE_ITHREADS */
+#endif /* DBI_NO_THREADS */
 
 
-/* forward declaration of 'DBI Handle Common Data', see below		*/
+/* forward struct declarations						*/
 
+typedef struct dbistate_st dbistate_t;
 /* implementor needs to define actual struct { dbih_??c_t com; ... }*/
 typedef struct imp_drh_st imp_drh_t;	/* driver			*/
 typedef struct imp_dbh_st imp_dbh_t;	/* database			*/
@@ -91,15 +88,15 @@ typedef struct dbih_com_std_st {
     HV   *my_h;		/* copy of outer handle HV (not refcounted)	*/
     SV   *parent_h;	/* parent inner handle (ref to hv) (r.c.inc)	*/
     imp_xxh_t *parent_com;	/* parent com struct shortcut		*/
-    dbi_cond  *thr_cond;/* condition for thread access (see dispatch)	*/
+    PerlInterpreter * thr_user;  /* thread that owns the handle         */
 
     HV   *imp_stash;	/* who is the implementor for this handle	*/
     SV   *imp_data;	/* optional implementors data (for perl imp's)	*/
 
     I32  kids;		/* count of db's for dr's, st's for db's etc	*/
     I32  active_kids;	/* kids which are currently DBIc_ACTIVE		*/
-    U32  thr_user;	/* thread id currently using the handle		*/
-    void *spare;
+    U32 pad;		/* keep binary compat */
+    dbistate_t *dbistate;
 } dbih_com_std_t;
 
 typedef struct dbih_com_attr_st {
@@ -109,7 +106,7 @@ typedef struct dbih_com_attr_st {
     SV *State;		/* Standard SQLSTATE, 5 char string	*/
     SV *Err;		/* Native engine error code		*/
     SV *Errstr;		/* Native engine error message		*/
-    SV *Handlers;
+    SV *spare;
     U32  LongReadLen;	/* auto read length for long/blob types	*/
     SV *FetchHashKeyName;	/* for fetchrow_hashref		*/
     /* (NEW FIELDS?... DON'T FORGET TO UPDATE dbih_clearcom()!)	*/
@@ -190,6 +187,8 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBIc_THR_USER_NONE  	(0xFFFF)
 #define DBIc_IMP_STASH(imp)  	_imp2com(imp, std.imp_stash)
 #define DBIc_IMP_DATA(imp)  	_imp2com(imp, std.imp_data)
+#define DBIc_DBISTATE(imp)  	_imp2com(imp, std.dbistate)
+#define DBIc_LOGPIO(imp)  	DBIc_DBISTATE(imp)->logfp
 #define DBIc_KIDS(imp)  	_imp2com(imp, std.kids)
 #define DBIc_ACTIVE_KIDS(imp)  	_imp2com(imp, std.active_kids)
 #define DBIc_LAST_METHOD(imp)  	_imp2com(imp, std.last_method)
@@ -199,7 +198,6 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBIc_STATE(imp)		SvRV(_imp2com(imp, attr.State))
 #define DBIc_ERR(imp)		SvRV(_imp2com(imp, attr.Err))
 #define DBIc_ERRSTR(imp)	SvRV(_imp2com(imp, attr.Errstr))
-#define DBIc_HANDLERS(imp)	SvRV(_imp2com(imp, attr.Handlers))
 #define DBIc_LongReadLen(imp)  	_imp2com(imp, attr.LongReadLen)
 #define DBIc_LongReadLen_init	80	/* may change */
 #define DBIc_FetchHashKeyName(imp) (_imp2com(imp, attr.FetchHashKeyName))
@@ -216,23 +214,26 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBIc_FDESC_AV(imp)  	_imp2com(imp, fields_fdav)
 #define DBIc_FDESC(imp, i)  	((imp_fdh_t*)(void*)SvPVX(AvARRAY(DBIc_FDESC_AV(imp))[i]))
 
+/* XXX --- DO NOT CHANGE THESE VALUES AS THEY ARE COMPILED INTO DRIVERS --- XXX */
 #define DBIcf_COMSET	  0x000001	/* needs to be clear'd before free'd	*/
 #define DBIcf_IMPSET	  0x000002	/* has implementor data to be clear'd	*/
 #define DBIcf_ACTIVE	  0x000004	/* needs finish/disconnect before clear	*/
 #define DBIcf_IADESTROY	  0x000008	/* do DBIc_ACTIVE_off before DESTROY	*/
 #define DBIcf_WARN  	  0x000010	/* warn about poor practice etc  	*/
 #define DBIcf_COMPAT  	  0x000020	/* compat/emulation mode (eg oraperl)	*/
-
 #define DBIcf_ChopBlanks  0x000040	/* rtrim spaces from fetch char columns	*/
 #define DBIcf_RaiseError  0x000080	/* throw exception (croak) on error	*/
 #define DBIcf_PrintError  0x000100	/* warn() on error			*/
 #define DBIcf_AutoCommit  0x000200	/* dbh only. used by drivers		*/
 #define DBIcf_LongTruncOk 0x000400	/* truncation to LongReadLen is okay	*/
 #define DBIcf_MultiThread 0x000800	/* allow multiple threads to enter	*/
-#define DBIcf_Taint       0x001000	/* taint fetched data			*/
-#define DBIcf_ShowErrorStatement  0x002000	/* include Statement in error	*/
+/*	spare		  0x001000						*/
+#define DBIcf_ShowErrorStatement  0x002000   /* include Statement in error	*/
 #define DBIcf_BegunWork   0x004000	/* between begin_work & commit/rollback */
-#define DBIcf_HandleError 0x008000	/* has coderef in HandleError attribute	*/
+#define DBIcf_HandleError 0x008000	/* has coderef in HandleError attribute */
+#define DBIcf_Profile     0x010000	/* profile activity on this handle      */
+#define DBIcf_TaintIn     0x020000	/* check inputs for taintedness */
+#define DBIcf_TaintOut    0x040000	/* taint outgoing data */
 
 #define DBIcf_INHERITMASK		/* what NOT to pass on to children */	\
   (U32)( DBIcf_COMSET | DBIcf_IMPSET | DBIcf_ACTIVE | DBIcf_IADESTROY		\
@@ -319,15 +320,14 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 #define DBI_IMP_SIZE(n,s) sv_setiv(perl_get_sv((n), GV_ADDMULTI), (s)) /* XXX */
 
 
-/* --- Implementors Field Descriptor Support --- */
-
 
 /* --- Event Support (VERY LIABLE TO CHANGE) --- */
 
-#define DBIh_EVENTx(h,t,a1,a2)	(DBIS->event((h), (t), (a1), (a2)))
-#define DBIh_EVENT0(h,t)	DBIh_EVENTx((h), (t), &sv_undef, &sv_undef)
-#define DBIh_EVENT1(h,t, a1)	DBIh_EVENTx((h), (t), (a1),      &sv_undef)
-#define DBIh_EVENT2(h,t, a1,a2)	DBIh_EVENTx((h), (t), (a1),      (a2))
+/* #define DBIh_EVENTx(h,t,a1,a2) (DBIS->event((h), (t), (a1), (a2))) */
+#define DBIh_EVENTx(h,t,a1,a2)	/* deprecated */ &PL_sv_no
+#define DBIh_EVENT0(h,t)	DBIh_EVENTx((h), (t), &PL_sv_undef, &PL_sv_undef)
+#define DBIh_EVENT1(h,t, a1)	DBIh_EVENTx((h), (t), (a1),         &PL_sv_undef)
+#define DBIh_EVENT2(h,t, a1,a2)	DBIh_EVENTx((h), (t), (a1),         (a2))
 
 #define ERROR_event	"ERROR"
 #define WARN_event	"WARN"
@@ -347,7 +347,7 @@ typedef struct {		/* -- FIELD DESCRIPTOR --		*/
 
 /* --- DBI State Structure --- */
 
-typedef struct {
+struct dbistate_st {
 
 #define DBISTATE_VERSION  94	/* Must change whenever dbistate_t does	*/
 
@@ -378,14 +378,15 @@ typedef struct {
     int         (*hash)		_((char *string, long i));
     SV        * (*preparse)	_((SV *sth, char *statement, IV ps_return, IV ps_accept, void *foo));
 
-    SV *neatsvpvlen;	/* only show dbgpvlen chars when debugging pv's	*/
+    SV *neatsvpvlen;		/* only show dbgpvlen chars when debugging pv's	*/
 
-    dbi_mutex	*mutex;
+    PerlInterpreter * thr_owner;	/* thread that owns this dbistate	*/
 
     int         (*logmsg)	_((imp_xxh_t *imp_xxh, char *fmt, ...));
     int         (*set_err)	_((imp_xxh_t *imp_xxh, char *fmt, ...));
-    void *pad[7];
-} dbistate_t;
+
+    void *pad2[7];
+};
 
 /* macros for backwards compatibility */
 #define set_attr(h, k, v)	set_attr_k(h, k, 0, v)
@@ -405,7 +406,7 @@ typedef struct {
 /* note that USE_ITHREADS implies MULTIPLICITY                      */
 #if defined(MULTIPLICITY) || defined(PERL_OBJECT) || defined(PERL_CAPI)
 
-# define DBISTATE_DECLARE
+# define DBISTATE_DECLARE typedef int dummy_dbistate /* keep semicolon from feeling lonely */
 # define DBISTATE_ASSIGN(st)
 # define DBISTATE_INIT
 static dbistate_t **get_dbistate() {
@@ -441,9 +442,10 @@ static dbistate_t **get_dbistate() {
 /* attribs value. One day we may add some extra magic in here.		*/
 #define DBD_ATTRIBS_CHECK(func, h, attribs)	\
     if ((attribs) && SvOK(attribs)) {		\
+	STRLEN lna1=0, lna2=0;			\
 	if (!SvROK(attribs) || SvTYPE(SvRV(attribs))!=SVt_PVHV)		\
 	    croak("%s->%s(...): attribute parameter '%s' is not a hash ref",	\
-		    SvPV(h,na), func, SvPV(attribs,na));		\
+		    SvPV(h,lna1), func, SvPV(attribs,lna2));		\
     } else (attribs) = Nullsv
 
 #define DBD_ATTRIB_GET_SVP(attribs, key,klen)			\
@@ -451,13 +453,24 @@ static dbistate_t **get_dbistate() {
 	    ? hv_fetch((HV*)SvRV(attribs), key,klen, 0)		\
 	    : (SV **)Nullsv)
 	
-#define DBD_ATTRIB_GET_BOOL(attribs, key,klen, svp, var)		\
-	if ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
-	    var = SvTRUE(*svp)
-
 #define DBD_ATTRIB_GET_IV(attribs, key,klen, svp, var)			\
 	if ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
 	    var = SvIV(*svp)
 
+#define DBD_ATTRIB_GET_BOOL(attribs, key,klen, svp, var)		\
+	if ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
+	    var = SvTRUE(*svp)
+
+#define DBD_ATTRIB_TRUE(attribs, key,klen, svp)				\
+	(  ((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)	\
+	    ? SvTRUE(*svp) : 0 )
+
+#define DBD_ATTRIB_GET_PV(attribs, key,klen, svp, dflt)			\
+	(((svp=DBD_ATTRIB_GET_SVP(attribs, key,klen)) != NULL)		\
+	    ? SvPV_nolen(*svp) : (dflt))
+
+#define DBD_ATTRIB_DELETE(attribs, key, klen)			\
+	hv_delete((HV*)attribs, key, len, G_DISCARD)
+
 #endif /* DBIXS_VERSION */
 /* end of DBIXS.h */
